@@ -34,6 +34,10 @@ def blob_oid(repo, changeset_id, path):
     return concat_chgsid_path(changeset_id, path)
 
 
+def ctx_blob_oid(changectx, path):
+    return blob_oid(changectx.repo(), changectx.hex().decode('ascii'), path)
+
+
 def extract_tree_oid(repo, oid: str) -> Tuple[str, Union[bytes, None]]:
     """"Return changeset id and path for a tree or commit OID.
 
