@@ -5,24 +5,12 @@
 #
 # SPDX-License-Identifier: GPL-2.0-or-later
 
-import logging
 from io import BytesIO
-import attr
+import logging
 
-from ..errors import (
-    not_found,
-    unknown_error,
-    unavailable_error,
-)
-from ..oid import (
-    blob_oid,
-)
-from ..git import (
-    NULL_BLOB_OID,
-    EMPTY_TREE_OID,
-)
+import attr
+from grpc import StatusCode
 from mercurial import (
-    copies,
     node,
     patch as patchmod,
     pycompat,
@@ -30,16 +18,40 @@ from mercurial import (
     cmdutil,
     scmutil,
 )
+
+from ..diff import (
+    Chunk,
+    changed_paths,
+    chunk_stats,
+    diff_opts,
+    git_patch_id,
+)
+from ..git import (
+    EMPTY_TREE_OID,
+)
+from ..logging import LoggerAdapter
+from ..oid import (
+    split_chgsid_path,
+)
+from ..revision import gitlab_revision_changeset
+from ..servicer import HGitalyServicer
+from ..stream import (
+    aggregate_flush_batch,
+    concat_resplit,
+    iter_boolean_lookahead,
+    split_batches,
+    WRITE_BUFFER_SIZE,
+)
 from ..stub.diff_pb2 import (
-    ChangedPaths,
     CommitDiffRequest,
     CommitDiffResponse,
     CommitDelta as MsgCommitDelta,
     CommitDeltaRequest,
     CommitDeltaResponse,
-    DiffStats as MsgDiffStats,
     FindChangedPathsRequest,
     FindChangedPathsResponse,
+    GetPatchIDRequest,
+    GetPatchIDResponse,
     RawDiffRequest,
     RawDiffResponse,
     RawPatchRequest,
@@ -48,20 +60,10 @@ from ..stub.diff_pb2 import (
     DiffStatsResponse,
 )
 from ..stub.diff_pb2_grpc import DiffServiceServicer
-
-from ..servicer import HGitalyServicer
-from ..revision import gitlab_revision_changeset
-from ..stream import (
-    aggregate_flush_batch,
-    concat_resplit,
-    iter_boolean_lookahead,
-    split_batches,
-    WRITE_BUFFER_SIZE,
-)
 from ..util import chunked
 
 
-logger = logging.getLogger(__name__)
+base_logger = logging.getLogger(__name__)
 # Copied from mercurial/patch.py
 # TODO even though, this is bytes and not octal, would be better
 # to use the constants in `hgitaly.git`
@@ -98,15 +100,6 @@ MAX_PATCH_BYTES_UPPERBOUND = 512000  # 500KB
 """MAX_PATCH_BYTES_UPPERBOUND controls how much max_patch_bytes limit can reach
 """
 
-Status_Type_Map = dict(
-    added=ChangedPaths.Status.ADDED,
-    modified=ChangedPaths.Status.MODIFIED,
-    removed=ChangedPaths.Status.DELETED,
-    copied=ChangedPaths.Status.COPIED,
-    # Note: Mercurial includes TYPE_CHANGE
-    # (symlink, regular file, submodule...etc) in MODIFIED status
-)
-
 
 @attr.s
 class Limits:
@@ -261,6 +254,18 @@ def get_commitDiff_parser(request):
     return Parser(limits, CurrDiff())
 
 
+WhiteSpaceChanges = CommitDiffRequest.WhitespaceChanges
+WHITESPACE_OPTS_MAPPING = {
+    WhiteSpaceChanges.WHITESPACE_CHANGES_IGNORE: b'ignore_space_change',
+    WhiteSpaceChanges.WHITESPACE_CHANGES_IGNORE_ALL: b'ignore_all_space',
+}
+"""Maps whitespace_changes in DiffRequest to opts for `hg diff`.
+
+Yes, it turns out that options of `hg diff` and `git diff` have exactly
+the same name.
+"""
+
+
 class DiffServicer(DiffServiceServicer, HGitalyServicer):
     """DiffService implementation.
 
@@ -274,14 +279,17 @@ class DiffServicer(DiffServiceServicer, HGitalyServicer):
         parsed_request = parse_diff_request(self, request, context)
         parsed, repo, ctx_from, ctx_to = parsed_request
         if not parsed:
-            return unavailable_error(context, CommitDiffResponse,
-                                     "CommitDiff: exit status 128")
+            context.abort(StatusCode.FAILED_PRECONDITION, "exit status 128")
+
         m = None
         if request.paths:
             m = scmutil.matchfiles(repo, request.paths)
         parser = get_commitDiff_parser(request)
         opts = {b'git': True}
-        whitespace = request.ignore_whitespace_change
+        # tell to interpret whitespace
+        whitespace = bool(request.whitespace_changes)
+        if whitespace:
+            opts[WHITESPACE_OPTS_MAPPING[request.whitespace_changes]] = True
         # XXX: add support and tests for request.ignore_whitespace_change
         diffopts = diffutil.difffeatureopts(
             repo.ui,
@@ -295,9 +303,10 @@ class DiffServicer(DiffServiceServicer, HGitalyServicer):
             opts=diffopts
         )
         for header in patchmod.parsepatch(diffchunks):
-            from_path, to_path = old_new_file_path(header)
-            from_id, to_id = old_new_blob_oids(header, ctx_from, ctx_to)
-            old_mode, new_mode = old_new_file_mode(header, ctx_from, ctx_to)
+            chunk = Chunk(header, ctx_from, ctx_to)
+            from_path, to_path = chunk.from_to_file_paths
+            from_id, to_id = chunk.from_to_blob_oids()
+            old_mode, new_mode = chunk.from_to_file_mode()
             # For CommitDiffResponse, modes are returned in decimal form
             old_mode, new_mode = int(old_mode, 8), int(new_mode, 8)
 
@@ -345,8 +354,7 @@ class DiffServicer(DiffServiceServicer, HGitalyServicer):
         parsed_request = parse_diff_request(self, request, context)
         _parsed, repo, ctx_from, ctx_to = parsed_request
         if not _parsed:
-            return unavailable_error(context, CommitDiffResponse,
-                                     "CommitDelta: exit status 128")
+            context.abort(StatusCode.FAILED_PRECONDITION, "exit status 128")
         m = None
         if request.paths:
             m = scmutil.matchfiles(repo, request.paths)
@@ -360,10 +368,10 @@ class DiffServicer(DiffServiceServicer, HGitalyServicer):
 
         def in_deltas():
             for header in patchmod.parsepatch(diffchunks):
-                from_path, to_path = old_new_file_path(header)
-                from_id, to_id = old_new_blob_oids(header, ctx_from, ctx_to)
-                old_mode, new_mode = old_new_file_mode(header, ctx_from,
-                                                       ctx_to)
+                chunk = Chunk(header, ctx_from, ctx_to)
+                from_path, to_path = chunk.from_to_file_paths
+                from_id, to_id = chunk.from_to_blob_oids()
+                old_mode, new_mode = chunk.from_to_file_mode()
                 # For CommitDeltaResponse, modes are returned in decimal form
                 old_mode, new_mode = int(old_mode, 8), int(new_mode, 8)
                 # As per Gitaly/Git behavior, if current Delta is a Rename and
@@ -393,7 +401,7 @@ class DiffServicer(DiffServiceServicer, HGitalyServicer):
         parsed_request = parse_diff_request(self, request, context)
         _parsed, repo, ctx_from, ctx_to = parsed_request
         if not _parsed:
-            return unknown_error(context, RawDiffResponse, "exit status 128")
+            context.abort(StatusCode.INTERNAL, "exit status 128")
         opts = {b'git': True}
         overrides = {
             (b'experimental', b'extendedheader.similarity'): True,
@@ -404,10 +412,12 @@ class DiffServicer(DiffServiceServicer, HGitalyServicer):
 
         # generator func to yield hunks
         def in_chunks():
-            for chunk in patchmod.parsepatch(diffchunks):
-                header = _insert_blob_index(chunk, ctx_from, ctx_to)
+            for hg_chunk in patchmod.parsepatch(diffchunks):
+                chunk = Chunk(hg_chunk, ctx_from, ctx_to)
+                header, _bin_placeholder = chunk.header_with_index_line()
                 yield header
-                for hunk in chunk.hunks:
+
+                for hunk in hg_chunk.hunks:
                     with BytesIO() as extracted:
                         hunk.write(extracted)
                         yield extracted.getvalue()
@@ -425,7 +435,7 @@ class DiffServicer(DiffServiceServicer, HGitalyServicer):
         parsed_request = parse_diff_request(self, request, context)
         _parsed, repo, ctx_from, ctx_to = parsed_request
         if not _parsed:
-            return unknown_error(context, RawPatchResponse, "exit status 128")
+            context.abort(StatusCode.INTERNAL, "cmd: exit status 128")
         opts = {b'git': True}
         diffopts = diffutil.diffallopts(repo.ui, opts)
         ui = repo.ui
@@ -446,70 +456,117 @@ class DiffServicer(DiffServiceServicer, HGitalyServicer):
         parsed_request = parse_diff_request(self, request, context)
         _parsed, repo, ctx_from, ctx_to = parsed_request
         if not _parsed:
-            return unavailable_error(context, DiffStatsResponse,
-                                     "DiffStats: exit status 128")
-        opts = {b'git': True}
-        diffopts = diffutil.difffeatureopts(repo.ui, opts=opts, git=True)
-        diffchunks = ctx_to.diff(
-            ctx_from,
-            opts=diffopts
-        )
+            context.abort(StatusCode.FAILED_PRECONDITION, "exit status 128")
 
-        def in_stats():
-            for header in patchmod.parsepatch(diffchunks):
-                old_path, path = old_new_file_path(header)
-                if old_path == path:
-                    old_path = b''
-                adds, dels = additions_deletions(header)
-                stats = MsgDiffStats(
-                    path=path,
-                    old_path=old_path,
-                    additions=adds,
-                    deletions=dels,
-                )
-                yield stats
-        for stats in aggregate_flush_batch(in_stats(), lambda x: 1,
-                                           MAX_NUM_STAT_BATCH_SIZE):
+        diffchunks = ctx_to.diff(ctx_from, opts=diff_opts(repo))
+
+        for stats in aggregate_flush_batch(
+                chunk_stats(diffchunks, ctx_from, ctx_to),
+                lambda x: 1,
+                MAX_NUM_STAT_BATCH_SIZE):
             yield DiffStatsResponse(stats=stats)
 
     def FindChangedPaths(self, request: FindChangedPathsRequest,
                          context) -> FindChangedPathsResponse:
         repo = self.load_repo(request.repository, context)
-        csets = []
-        for cid in request.commits:
-            ctx = gitlab_revision_changeset(repo, parse_diff_request_cid(cid))
-            if ctx is None:
-                yield not_found(
-                    context, FindChangedPathsResponse,
-                    'FindChangedPaths: commit: %s can not be '
-                    'found' % pycompat.sysstr(cid)
-                )
-                return
-            csets.append(ctx)
+        extracted = []  # triplets (changeset1, changeset2, path)
+        SubReq = FindChangedPathsRequest.Request
+        CommitRequest = SubReq.CommitRequest
 
-        def in_changed_paths():
-            for ctx in csets:
-                p1 = ctx.p1()
-                status = p1.status(ctx)
-                for path in _status_paths(status, Status_Type_Map):
-                    yield path
-            copy = copies.pathcopies(p1, ctx)
-            for path in copy:
-                yield ChangedPaths(
-                    path=path,
-                    status=Status_Type_Map['copied']
-                )
-        for paths in chunked(in_changed_paths()):
+        sub_reqs = request.requests
+        if request.commits:
+            # compat wrapper for deprecated call style. Quoting diff.proto:
+            #   This field is deprecated. To adapt to the new calling
+            #   convention you can create one `CommitRequest` per commit,
+            #   where each `CommitRequest` has only the `commit_revision`
+            sub_reqs.extend(
+                SubReq(commit_request=CommitRequest(commit_revision=rev))
+                for rev in request.commits
+            )
+
+        # Gitaly (as of gitaly@2b069d853) simply issues a `git diff-tree`
+        # per sub-request and yields the responses of each.
+        # In other words, there is no merging of results. For instance,
+        # with two sub-requests, a file appearing modified for both of them
+        # will give rise to two identical ChangedPaths messages.
+        for sub_req in sub_reqs:
+            if sub_req.HasField('commit_request'):
+                right_ctx = fcp_resolve_commit(
+                    context, repo,
+                    sub_req.commit_request.commit_revision)
+
+                parent_revs = sub_req.commit_request.parent_commit_revisions
+                if parent_revs:
+                    parents = [fcp_resolve_commit(context, repo, rev)
+                               for rev in parent_revs]
+                else:
+                    parents = right_ctx.parents()
+
+                extracted.extend((left_ctx, right_ctx, None)
+                                 for left_ctx in parents)
+            else:
+                # TODO trees can probably be specified in some arcane Git
+                # traversal rather than just by oid.
+                oids = [sub_req.tree_request.left_tree_revision,
+                        sub_req.tree_request.right_tree_revision]
+                cids_paths = [split_chgsid_path(oid) for oid in oids]
+                path = cids_paths[0][1]
+                if cids_paths[1][1] != path:
+                    context.abort(StatusCode.UNIMPLEMENTED,
+                                  "Cannot compare yet with different paths. "
+                                  "Got TreeRequest with left and right "
+                                  "resolving as %r" % cids_paths)
+                unfi = repo.unfiltered()
+                left_ctx, right_ctx = [unfi[cp[0].encode('ascii')]
+                                       for cp in cids_paths]
+                extracted.append((left_ctx, right_ctx, path))
+
+        find_renames = request.find_renames
+        for paths in chunked(path for extr in extracted
+                             for path in changed_paths(
+                                     repo, *extr,
+                                     find_renames=find_renames)):
             yield FindChangedPathsResponse(paths=paths)
 
+    def GetPatchID(self, request: GetPatchIDRequest,
+                   context) -> GetPatchIDResponse:
+        repo = self.load_repo(request.repository, context)
+        old_changeset = gitlab_revision_changeset(repo, request.old_revision)
+        if old_changeset is None:
+            context.abort(StatusCode.INTERNAL,
+                          "revision %r not found" % request.old_revision)
+        new_changeset = gitlab_revision_changeset(repo, request.new_revision)
+        if new_changeset is None:
+            context.abort(StatusCode.INTERNAL,
+                          "revision %r not found" % request.new_revision)
 
-def _status_paths(status, type_map):
-    for stype in ['added', 'modified', 'removed']:
-        for path in status.__getattribute__(stype):
-            yield ChangedPaths(
-                path=path,
-                status=type_map[stype]
-            )
+        git_path = repo.ui.config(b'hgitaly', b'git-executable')
+
+        try:
+            git_out = git_patch_id(git_path, old_changeset, new_changeset)
+        except FileNotFoundError:
+            context.abort(StatusCode.INTERNAL,
+                          "Expected Git executable not found at %r" % git_path)
+        except PermissionError:
+            context.abort(StatusCode.INTERNAL,
+                          "Expected Git executable found at %r, but it is "
+                          "not executable" % git_path)
+
+        # like Gitaly, ignoring the second hash, which is useful only
+        # in Git tree comparisons (to recall a Git commit id)
+        return GetPatchIDResponse(patch_id=git_out.split()[0])
+
+
+def fcp_resolve_commit(context, repo, revision):
+    ctx = gitlab_revision_changeset(
+        repo, parse_diff_request_cid(revision))
+    if ctx is None:
+        context.abort(
+            StatusCode.NOT_FOUND,
+            'resolving commit: revision can not be found: '
+            '"%s"' % pycompat.sysstr(revision)
+        )
+    return ctx
 
 
 def parse_diff_request_cid(cid):
@@ -521,6 +578,7 @@ def parse_diff_request_cid(cid):
 
 
 def parse_diff_request(servicer, request, context):
+    logger = LoggerAdapter(base_logger, context)
     repo = servicer.load_repo(request.repository, context)
 
     left_cid = parse_diff_request_cid(request.left_commit_id)
@@ -541,73 +599,6 @@ def parse_diff_request(servicer, request, context):
     return (True, repo, ctx_from, ctx_to)
 
 
-def old_new_blob_oids(header, old_ctx, new_ctx):
-    """Return a tuple of (old, new) blob oids."""
-    old_path, new_path = old_new_file_path(header)
-    old_bid = new_bid = NULL_BLOB_OID
-    if old_path in old_ctx:
-        cid = pycompat.sysstr(old_ctx.hex())
-        old_bid = blob_oid(None, cid, old_path)
-    if new_path in new_ctx:
-        cid = pycompat.sysstr(new_ctx.hex())
-        new_bid = blob_oid(None, cid, new_path)
-    return old_bid, new_bid
-
-
-def old_new_file_mode(header, old_ctx, new_ctx):
-    """Return a tuple of (old, new) file mode."""
-    old_path, new_path = old_new_file_path(header)
-    old_mode, new_mode = b'0', b'0'
-    if old_path in old_ctx:
-        old_fctx = old_ctx[old_path]
-        old_mode = gitmode[old_fctx.flags()]
-    if new_path in new_ctx:
-        new_fctx = new_ctx[new_path]
-        new_mode = gitmode[new_fctx.flags()]
-    return old_mode, new_mode
-
-
-def old_new_file_path(header):
-    """Return a tuple of (old, new) file path."""
-    fname = header.filename()
-    from_path, to_path = fname, fname
-    if len(header.files()) > 1:
-        # file is renamed
-        from_path, to_path = header.files()
-    return from_path, to_path
-
-
-def additions_deletions(header):
-    adds, dels = 0, 0
-    for hunk in header.hunks:
-        add_count, del_count = hunk.countchanges(hunk.hunk)
-        adds += add_count
-        dels += del_count
-    return adds, dels
-
-
-def _insert_blob_index(chunk, ctx_from, ctx_to):
-    fname = chunk.filename()
-    old_bid, new_bid = old_new_blob_oids(chunk, ctx_from, ctx_to)
-    indexline = 'index %s..%s' % (old_bid, new_bid)
-    indexline = pycompat.sysbytes(indexline)
-
-    # Note: <mode> is required only when it didn't change between
-    # the two changesets, otherwise it has a separate line
-    if fname in ctx_from and fname in ctx_to:
-        oldmode, mode = old_new_file_mode(chunk, ctx_from, ctx_to)
-        if mode == oldmode:
-            indexline += b' ' + mode
-    indexline += b'\n'
-    headerlines = chunk.header
-
-    for index, line in enumerate(headerlines[:]):
-        if line.startswith(b'--- '):
-            headerlines.insert(index, indexline)
-            break
-    return b''.join(headerlines)
-
-
 def _exportsingle(repo, ctx, fm, seqno, diffopts):
     """Generator method which yields a bytes stream of exporting `ctx` data.
 
