@@ -5,15 +5,20 @@
 #
 # SPDX-License-Identifier: GPL-2.0-or-later
 import logging
+from grpc import StatusCode
 
 from mercurial import (
-    pycompat,
+    error as hgerror,
 )
 
 from heptapod.gitlab.branch import (
+    GITLAB_BRANCH_REF_PREFIX,
     gitlab_branch_from_ref,
     gitlab_branch_ref,
 )
+from heptapod.gitlab.tag import (
+    GITLAB_TAG_REF_PREFIX,
+)
 from hgext3rd.heptapod.branch import get_default_gitlab_branch
 from hgext3rd.heptapod.typed_ref import GITLAB_TYPED_REFS_MISSING
 from hgext3rd.heptapod.special_ref import (
@@ -23,35 +28,47 @@ from hgext3rd.heptapod.special_ref import (
 )
 from hgext3rd.heptapod.keep_around import (
     parse_keep_around_ref,
+    iter_keep_arounds,
+    init_keep_arounds,
 )
 
 from ..errors import (
-    invalid_argument,
-    not_found,
     not_implemented,
+    structured_abort,
 )
+from ..logging import LoggerAdapter
 from ..pagination import (
     extract_limit,
 )
+from ..revision import (
+    CHANGESET_HASH_BYTES_REGEXP,
+    ZERO_SHA,
+    ZERO_SHA_STR,
+    gitlab_revision_hash,
+)
+from ..stream import (
+    chunked_limit,
+)
 from ..stub.shared_pb2 import (
     Branch,
-    GitCommit,
+    SortDirection,
+)
+from ..stub.errors_pb2 import (
+    ReferenceNotFoundError,
+    ReferenceStateMismatchError,
 )
 from ..stub.ref_pb2 import (
     FindDefaultBranchNameRequest,
     FindDefaultBranchNameResponse,
-    FindAllBranchNamesRequest,
-    FindAllBranchNamesResponse,
-    FindAllTagNamesRequest,
-    FindAllTagNamesResponse,
     FindLocalBranchesRequest,
-    FindLocalBranchCommitAuthor,
-    FindLocalBranchResponse,
     FindLocalBranchesResponse,
     FindAllBranchesRequest,
     FindAllBranchesResponse,
     FindAllTagsRequest,
     FindAllTagsResponse,
+    FindRefsByOIDRequest,
+    FindRefsByOIDResponse,
+    FindTagError,
     FindTagRequest,
     FindTagResponse,
     FindAllRemoteBranchesRequest,
@@ -62,6 +79,9 @@ from ..stub.ref_pb2 import (
     FindBranchResponse,
     DeleteRefsRequest,
     DeleteRefsResponse,
+    UpdateReferencesError,
+    UpdateReferencesRequest,
+    UpdateReferencesResponse,
     ListBranchNamesContainingCommitRequest,
     ListBranchNamesContainingCommitResponse,
     ListTagNamesContainingCommitRequest,
@@ -70,12 +90,8 @@ from ..stub.ref_pb2 import (
     GetTagSignaturesResponse,
     GetTagMessagesRequest,
     GetTagMessagesResponse,
-    PackRefsRequest,
-    PackRefsResponse,
     ListRefsRequest,
     ListRefsResponse,
-    FindRefsByOIDRequest,
-    FindRefsByOIDResponse,
 )
 from ..stub.ref_pb2_grpc import RefServiceServicer
 
@@ -89,23 +105,36 @@ from ..branch import (
 from ..gitlab_ref import (
     ensure_special_refs,
     has_keep_around,
+    iter_gitlab_special_refs_as_refs,
+    iter_keep_arounds_as_refs,
 )
 from ..tag import (
+    EXCLUDED_TAG_TYPES,
+    iter_gitlab_tags,
     iter_gitlab_tags_as_refs,
 )
 from .. import message
 from ..servicer import HGitalyServicer
 from ..util import chunked
 
-logger = logging.getLogger(__name__)
+base_logger = logging.getLogger(__name__)
 DEFAULT_BRANCH_FILE_NAME = b'default_gitlab_branch'
-EXCLUDED_TAG_TYPES = (b'local', None)  # 'tip' has type None
 FIND_LOCAL_BRANCHES_SORT_BY_TO_INNER = {
     FindLocalBranchesRequest.SortBy.NAME: BranchSortBy.FULL_REF_NAME,
     FindLocalBranchesRequest.SortBy.UPDATED_ASC: BranchSortBy.UPDATED_ASC,
     FindLocalBranchesRequest.SortBy.UPDATED_DESC: BranchSortBy.UPDATED_DESC
 }
 
+FIND_REFS_BY_OID_SORT_FIELD_TO_LIST_REFS = dict(
+    refname=ListRefsRequest.SortBy.REFNAME,
+    creatordate=ListRefsRequest.SortBy.CREATORDATE,
+    authordate=ListRefsRequest.SortBy.AUTHORDATE,
+)
+"""keys are amenable fields in git-for-each-ref(1)
+
+Not many such keys actually map to a SortBy value for ListRefsRequest.
+"""
+
 
 class RefServicer(RefServiceServicer, HGitalyServicer):
     """RefService implementation.
@@ -117,53 +146,27 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
             self,
             request: FindDefaultBranchNameRequest,
             context) -> FindDefaultBranchNameResponse:
-        logger.debug("Processing %r", message.Logging(request))
-        try:
-            repo = self.load_repo(request.repository, context)
-        except KeyError as exc:
-            return not_found(context, FindDefaultBranchNameResponse,
-                             "repository not found: " + repr(exc.args))
+        logger = LoggerAdapter(base_logger, context)
+        repo = self.load_repo(request.repository, context)
+
         branch = get_default_gitlab_branch(repo)
-        if branch is None:
+        if branch is None or gitlab_branch_head(repo, branch) is None:
             # Very coincidental, but this ends up as empty GitLab branch name
             # which the chain of || in app/models/concerns/repository
             # that eventually turns this in `nil`, which is the marker
             # used in the hg_fixup_default_branch in PostReceive…
             # TODO now that we have richer notifications, make a clear one
             # for that case
-            branch = b''
             logger.warning("FindDefaultBranchName: no information stored "
-                           "for repo at %r, returning harcoded default %r",
-                           repo, branch)
+                           "for repo at %r, returning hempty response",
+                           repo)
+            return FindDefaultBranchNameResponse()
 
         return FindDefaultBranchNameResponse(name=gitlab_branch_ref(branch))
 
-    def FindAllBranchNames(self,
-                           request: FindAllBranchNamesRequest,
-                           context) -> FindAllBranchNamesResponse:
-        logger.debug("Processing %r", message.Logging(request))
-        repo = self.load_repo(request.repository, context)
-        for chunk in chunked(iter_gitlab_branches_as_refs(repo)):
-            yield FindAllBranchNamesResponse(names=(br[0] for br in chunk))
-
-    def FindAllTagNames(self,
-                        request: FindAllTagNamesRequest,
-                        context) -> FindAllTagNamesResponse:
-        logger.debug("Processing %r", message.Logging(request))
-        repo = self.load_repo(request.repository, context)
-        # TODO SPEC it's not clear whether GitLab actually expect the
-        # tags to be in any ordering. At a quick glance it doesn't, but
-        # maye we'll end up with confused users because it'd make no sense
-        # in the UI.
-        tags = (t for t in repo.tags()
-                if repo.tagtype(t) not in EXCLUDED_TAG_TYPES)
-        for chunk in chunked(iter_gitlab_tags_as_refs(tags)):
-            yield FindAllTagNamesResponse(names=chunk)
-
     def FindLocalBranches(self,
                           request: FindLocalBranchesRequest,
                           context) -> FindLocalBranchesResponse:
-        logger.debug("Processing %r", message.Logging(request))
         repo = self.load_repo(request.repository, context)
         limit = extract_limit(request)
         if limit == 0:
@@ -175,22 +178,25 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
         # TODO encoding?
         after = after.encode('utf-8') if after else None
         sort_by = FIND_LOCAL_BRANCHES_SORT_BY_TO_INNER.get(request.sort_by)
+
         for chunk in chunked(sorted_gitlab_branches_as_refs(repo,
                                                             sort_by=sort_by,
                                                             after=after),
                              limit=limit):
             yield FindLocalBranchesResponse(
-                branches=(find_local_branch_response(name, head)
-                          for name, head in chunk),
+                local_branches=(message.branch(name, head)
+                                for name, head in chunk)
             )
 
     def FindAllBranches(self,
                         request: FindAllBranchesRequest,
                         context) -> FindAllBranchesResponse:
-        logger.debug("Processing %r", message.Logging(request))
         Branch = FindAllBranchesResponse.Branch
         repo = self.load_repo(request.repository, context)
-        for chunk in chunked(iter_gitlab_branches(repo)):
+        for chunk in chunked(sorted_gitlab_branches_as_refs(
+                repo,
+                sort_by=BranchSortBy.FULL_REF_NAME
+        )):
             yield FindAllBranchesResponse(
                 branches=(Branch(name=name, target=message.commit(head))
                           for name, head in chunk))
@@ -198,24 +204,21 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
     def FindAllTags(self,
                     request: FindAllTagsRequest,
                     context) -> FindAllTagsResponse:
-        logger.debug("Processing %r", message.Logging(request))
         repo = self.load_repo(request.repository, context)
-        for chunk in chunked(
-                item for item in pycompat.iteritems(repo.tags())
-                if repo.tagtype(item[0]) not in EXCLUDED_TAG_TYPES):
-            yield FindAllTagsResponse(tags=[message.tag(name, repo[node])
-                                            for name, node in chunk])
+        for chunk in chunked(iter_gitlab_tags(repo)):
+            yield FindAllTagsResponse(tags=[message.tag(name, ctx)
+                                            for name, ctx in chunk])
 
     def FindTag(self,
                 request: FindTagRequest,
                 context) -> FindTagResponse:
-        logger.debug("Processing %r", message.Logging(request))
         repo = self.load_repo(request.repository, context)
         name = request.tag_name
         if repo.tagtype(name) in EXCLUDED_TAG_TYPES:
-            # TODO SPEC confirm that this is the accepted response for
-            # not found tag.
-            return FindTagResponse(tag=None)
+            structured_abort(
+                context, StatusCode.NOT_FOUND, "tag does not exist",
+                FindTagError(tag_not_found=ReferenceNotFoundError(
+                    reference_name=b'refs/tags/' + name)))
 
         node = repo.tags()[name]
         return FindTagResponse(tag=message.tag(name, repo[node]))
@@ -224,19 +227,14 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
                               request: FindAllRemoteBranchesRequest,
                               context) -> FindAllRemoteBranchesResponse:
         """There is no concept of "remote branch" in Mercurial."""
-        logger.debug("Processing %r", message.Logging(request))
         return iter(())
 
     def RefExists(self,
                   request: RefExistsRequest,
                   context) -> RefExistsResponse:
-        logger.debug("Processing %r", message.Logging(request))
         ref = request.ref
         if not ref.startswith(b'refs/'):
-            # TODO SPEC commment in ref.proto says `ref` must start with
-            # refs/ but doesn't say if False should be returned or an error
-            # be set.
-            return RefExistsResponse(value=False)
+            context.abort(StatusCode.INVALID_ARGUMENT, "invalid refname")
 
         # TODO protect here
         repo = self.load_repo(request.repository, context)
@@ -267,7 +265,6 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
     def FindBranch(self,
                    request: FindBranchRequest,
                    context) -> FindBranchResponse:
-        logger.debug("Processing %r", message.Logging(request))
         repo = self.load_repo(request.repository, context)
         name = request.name
         if name.startswith(b'refs/'):
@@ -284,16 +281,101 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
         return FindBranchResponse(
             branch=Branch(name=name, target_commit=message.commit(head)))
 
+    def UpdateReferences(self,
+                         request: UpdateReferencesRequest,
+                         context) -> UpdateReferencesResponse:
+        self.STATUS_CODE_STORAGE_NOT_FOUND = StatusCode.INVALID_ARGUMENT
+        first = True
+        special_changes = []
+        ka_changes = []
+        for req in request:
+            if first:
+                # blobs are given by oid, hence as in direct changeset Node
+                # IDs, the unfiltered repo is the right one for the task
+                repo = self.load_repo(req.repository, context).unfiltered()
+                first = False
+
+            for upd in req.updates:
+                if len(upd.new_object_id) != 40:
+                    oid = upd.new_object_id.decode('ascii', 'replace')
+                    context.abort(StatusCode.INVALID_ARGUMENT,
+                                  "validating new object ID: "
+                                  f'invalid object ID: "{oid}", '
+                                  f"expected length 40, got {len(oid)}")
+
+                ref_path = upd.reference
+                change = (upd.old_object_id, upd.new_object_id)
+                if change[0] == change[1]:
+                    # let's not waste resources for this
+                    continue
+
+                # consider only special refs and keep-arounds, as other refs
+                # (branches and tags) are only reflecting changesets content,
+                # setting them independently makes no sense with Mercurial
+                special_ref = parse_special_ref(ref_path)
+                if special_ref is not None:
+                    special_changes.append((special_ref, change))
+                    continue
+
+                ka = parse_keep_around_ref(upd.reference)
+                if ka is not None:
+                    if (
+                            upd.new_object_id != ZERO_SHA
+                            and ka != upd.new_object_id
+                    ):
+                        context.abort(StatusCode.INVALID_ARGUMENT,
+                                      "Inconsistent keep-around.")
+                    ka_changes.append(change)
+
+        with repo.lock():
+            # keeping promise of atomicity of this method
+            special_existing = special_refs(repo)
+            if special_existing is GITLAB_TYPED_REFS_MISSING:
+                special_existing = {}
+            for name, (old_id, new_id) in special_changes:
+                existing = special_existing.get(name)
+                if old_id == ZERO_SHA and existing is not None:
+                    context.abort(StatusCode.INTERNAL,
+                                  "committing update: "
+                                  "reference already exists")
+                elif old_id and old_id != ZERO_SHA and existing != old_id:
+                    Mismatch = ReferenceStateMismatchError  # just too long!
+                    structured_abort(
+                        context, StatusCode.ABORTED,
+                        "reference does not point to expected object",
+                        UpdateReferencesError(
+                            reference_state_mismatch=Mismatch(
+                                reference_name=b'refs/' + name,
+                                expected_object_id=old_id,
+                                actual_object_id=existing)))
+                elif new_id == ZERO_SHA:
+                    del special_existing[name]
+                else:
+                    special_existing[name] = new_id
+
+            # TODO make a collect_keep_arounds() or keep_arounds_set()
+            ka_existing = set(iter_keep_arounds(repo))
+            ka_existing.discard(GITLAB_TYPED_REFS_MISSING)
+            for old_id, new_id in ka_changes:
+                # bogus requests have already been prohibited
+                if new_id == ZERO_SHA:
+                    ka_existing.discard(old_id)
+                else:
+                    ka_existing.add(new_id)
+
+            write_special_refs(repo, special_existing)
+            init_keep_arounds(repo, ka_existing)
+
+        return UpdateReferencesResponse()
+
     def DeleteRefs(self,
                    request: DeleteRefsRequest,
                    context) -> DeleteRefsResponse:
-        logger.debug("Processing %r", message.Logging(request))
         except_prefix = request.except_with_prefix
         refs = request.refs
         if refs and except_prefix:
-            return invalid_argument(
-                context, DeleteRefsResponse,
-                "DeleteRefs: ExceptWithPrefix and Refs are mutually exclusive")
+            context.abort(StatusCode.INVALID_ARGUMENT,
+                          "ExceptWithPrefix and Refs are mutually exclusive")
 
         repo = self.load_repo(request.repository, context)
         srefs = special_refs(repo)
@@ -319,7 +401,6 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
             self,
             request: ListBranchNamesContainingCommitRequest,
             context) -> ListBranchNamesContainingCommitResponse:
-        logger.debug("Processing %r", message.Logging(request))
         repo = self.load_repo(request.repository, context)
 
         gl_branches_by_heads = {}
@@ -327,6 +408,7 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
             rev = head.rev()
             gl_branches_by_heads.setdefault(rev, []).append(gl_branch)
 
+        repo = repo.unfiltered()
         heads = repo.revs(b'%ld and %s::', gl_branches_by_heads,
                           request.commit_id)
         # TODO SPEC since there's a limit, we'll have to know what is
@@ -349,10 +431,14 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
             self,
             request: ListTagNamesContainingCommitRequest,
             context) -> ListTagNamesContainingCommitResponse:
-        logger.debug("Processing %r", message.Logging(request))
         # TODO support ordering, see similar method for branches
         repo = self.load_repo(request.repository, context)
-        revs = repo.revs("%s:: and tag()", request.commit_id)
+
+        try:
+            revs = repo.revs("%s:: and tag()", request.commit_id)
+        except hgerror.FilteredRepoLookupError:
+            revs = []
+
         tag_names = (name for rev in revs for name in repo[rev].tags()
                      if repo.tagtype(name) not in EXCLUDED_TAG_TYPES)
         for chunk in chunked(tag_names, limit=request.limit):
@@ -361,8 +447,7 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
     def GetTagSignatures(self,
                          request: GetTagSignaturesRequest,
                          context) -> GetTagSignaturesResponse:
-        return not_implemented(context, GetTagSignaturesResponse,
-                               issue=75)  # pragma no cover
+        not_implemented(context, issue=75)  # pragma no cover
 
     def GetTagMessages(self,
                        request: GetTagMessagesRequest,
@@ -375,7 +460,6 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
         For now, we'll consider that the id of a tag is the nod id of the
         changeset that gives it its current value.
         """
-        logger.debug("Processing %r", message.Logging(request))
         repo = self.load_repo(request.repository, context)
         # TODO check that the given id is indeed for a tag, i.e. a
         # changeset that affects .hgtags?
@@ -434,53 +518,141 @@ class RefServicer(RefServiceServicer, HGitalyServicer):
         raise NotImplementedError(
             "Not relevant for Mercurial")  # pragma: no cover
 
-    def PackRefs(self, request: PackRefsRequest, context) -> PackRefsResponse:
-        """Not relevant for Mercurial, does nothing.
+    def iter_refs(self, repo, request: ListRefsRequest, context):
+        """"Generator yielding refernece names and target shas.
+
+        For very direct use in ListRefs and similar.
         """
-        # repr(Repository) contains newlines
-        logger.warning("Ignored irrelevant PackRefs request %r",
-                       message.Logging(request))
-        return PackRefsResponse()
+        patterns = request.patterns
+
+        refs = []
+        if patterns == [b'refs/'] and len(repo):
+            refs.append((b'ALL', ZERO_SHA_STR))
+        if request.head:
+            branch = get_default_gitlab_branch(repo)
+            if branch is not None:
+                head = gitlab_branch_head(repo, branch)
+                if head is not None:
+                    refs.append((b'HEAD', head.hex().decode('ascii')))
+        if patterns == [GITLAB_BRANCH_REF_PREFIX]:
+            refs.extend(iter_gitlab_branches_as_refs(repo, deref=False))
+        elif patterns == [GITLAB_TAG_REF_PREFIX]:
+            refs.extend(iter_gitlab_tags_as_refs(repo, deref=False))
+        else:
+            if patterns:
+                patterns = tuple(pat.split(b'/') for pat in patterns)
+            refs.extend(iter_gitlab_branches_as_refs(repo,
+                                                     deref=False,
+                                                     patterns=patterns))
+            refs.extend(iter_gitlab_tags_as_refs(repo,
+                                                 deref=False,
+                                                 patterns=patterns))
+            refs.extend(iter_gitlab_special_refs_as_refs(repo,
+                                                         deref=False,
+                                                         patterns=patterns))
+            refs.extend(iter_keep_arounds_as_refs(repo,
+                                                  deref=False,
+                                                  patterns=patterns))
+
+        pointing_at = request.pointing_at_oids
+        if pointing_at:
+            refs = [r for r in refs
+                    if r[1].encode('ascii') in pointing_at or r[0] == b'HEAD']
+
+        # ignore `peel_tags`:
+        # Mercurial has no real equivalent of Git's annotated tags. The
+        # closest notion would be the tagging changeset, but that is not
+        # well-defined. Hence Mercurial tags always point to the target commit.
+        # We could force the `peeled_target` attribute of the `Reference`
+        # messages, but that would not bring us to better compliance: the
+        # protocol comment states explicitely that it should stay empty if
+        # the object is not an annotated tag, and this is confirmed by
+        # Gitaly Comparison tests (tag created by hg-git in this case).
+        # On the contrary, filling-in `peeled_target` could very well make
+        # clients believe that the tag is annotated, leading to dedicated
+        # actions that can only go wrong.
+
+        if len(refs) < (2 if request.head else 1):
+            # avoid yielding an empty response, as Gitaly does not.
+            # TODO consider doing that in the `chunked()` helper,
+            # this looks to be rather systematic.
+            # also in this case, even if `head` is `True`, in the Gitaly
+            # impl, it looks like nothing being read from Git does not
+            # trigger any sending, even though the sender created by
+            # `newListRefsWriter` would emit HEAD first in any case… if it
+            # emitted anything. See Gitaly Comparison test for validation of
+            # this.
+            return
+
+        sort_by = request.sort_by
+        if sort_by.key == ListRefsRequest.SortBy.Key.REFNAME:
+            sort_key_fun = None
+        else:
+            # Some references can point to obsolete changesets
+            # (e.g, keep-arounds). In any case it is certainly not the
+            # role of this method to prevent one to do so.
+            unfi = repo.unfiltered()
+
+            # all other choices are about dates, and we have no other choice
+            # that conflating them with the commit date of the ref target
+            def sort_key_fun(ref_tuple):
+                sha = ref_tuple[1]
+                return unfi[sha].date()
+
+        refs.sort(key=sort_key_fun,
+                  reverse=sort_by.direction == SortDirection.DESCENDING)
+
+        for chunk in chunked(refs):
+            yield chunk
 
     def ListRefs(self, request: ListRefsRequest,
                  context) -> ListRefsResponse:
-        return not_implemented(context, ListRefsResponse,
-                               issue=90)  # pragma no cover
+        repo = self.load_repo(request.repository, context)
+        Reference = ListRefsResponse.Reference
+        for chunk in self.iter_refs(repo, request, context):
+            yield ListRefsResponse(
+                references=(Reference(name=name, target=target)
+                            for name, target in chunk))
 
     def FindRefsByOID(self, request: FindRefsByOIDRequest,
                       context) -> FindRefsByOIDResponse:
-        return not_implemented(context, FindRefsByOIDResponse,
-                               issue=89)  # pragma no cover
+        repo = self.load_repo(request.repository, context)
+        oid = request.oid.encode('utf8')
+        if not CHANGESET_HASH_BYTES_REGEXP.match(oid):
+            # a bit wider than needed, would probably be more
+            # efficient to query nodemap directly, but we'll consider this
+            # the day we reimplement in RHGitaly
+            oid = gitlab_revision_hash(repo, oid)
+        sort_by = FIND_REFS_BY_OID_SORT_FIELD_TO_LIST_REFS.get(
+            request.sort_field, ListRefsRequest.SortBy.REFNAME)
 
+        patterns = request.ref_patterns
+        if patterns:
+            patterns = [pat.encode('utf8') for pat in patterns]
+        else:
+            patterns = [b'refs/heads/', b'refs/tags/']
 
-def flbr_author(commit: GitCommit):
-    """Extract commit intro specific fields of FindLocalBranchCommitAuthor."""
-    return FindLocalBranchCommitAuthor(
-        name=commit.name,
-        email=commit.email,
-        date=commit.date,
-        timezone=commit.timezone,
-    )
+        limit = request.limit
+        if limit == 0:  # said to mean no limit in protocol comment
+            limit = None
 
-
-def find_local_branch_response(name, head):
-    commit = message.commit(head)
-    return FindLocalBranchResponse(
-        name=name,
-        commit_id=commit.id,
-        commit_subject=commit.subject,
-        commit_author=flbr_author(commit.author),
-        commit_committer=flbr_author(commit.committer),
-        commit=commit,
-    )
-
-
-TAG_REF_PREFIX = b'refs/tags/'
+        refs = []
+        for chunk in chunked_limit(self.iter_refs(
+                repo,
+                ListRefsRequest(
+                    repository=request.repository,
+                    pointing_at_oids=[oid],
+                    patterns=patterns,
+                    sort_by=ListRefsRequest.SortBy(key=sort_by),
+                ),
+                context), limit):
+            refs.extend(name for name, _tgt in chunk)
+        return FindRefsByOIDResponse(refs=refs)
 
 
 def gitlab_tag_from_ref(ref):
-    if ref.startswith(TAG_REF_PREFIX):
-        return ref[len(TAG_REF_PREFIX):]
+    if ref.startswith(GITLAB_TAG_REF_PREFIX):
+        return ref[len(GITLAB_TAG_REF_PREFIX):]
 
 
 def remove_special_refs(special_refs, to_remove):
