@@ -5,6 +5,9 @@
 #
 # SPDX-License-Identifier: GPL-2.0-or-later
 import pytest
+from importlib import reload
+
+from mercurial import sshpeer
 
 from mercurial_testhelpers.repo_wrapper import RepoWrapper
 from ..sshd import hg_ssh_setup
@@ -24,6 +27,7 @@ def test_ssh_server(ssh_setup, tmpdir):
     src = RepoWrapper.init(tmpdir / 'src')
     dest = RepoWrapper.init(tmpdir / 'dst')
     ctx = src.commit_file('foo')
+    reload(sshpeer)
 
     with client_command_str(client_key, known_hosts) as ssh_cmd:
         src.command('push',
