@@ -121,6 +121,9 @@ class WorkingDirectory:
         # this module is to provide a higher level locking system
         repo.vfs.tryunlink(b'wlock')
 
+    def file_path(self, relative_path):
+        return os.path.join(bytes(self.path), relative_path)
+
 
 @contextmanager
 def working_directory(workdirs_root, repo, *args, **kwargs):
