@@ -11,12 +11,16 @@ ensure that the foundation of all Gitaly comparison tests keeps working.
 """
 import pytest
 
-from hgitaly.stub.repository_service_pb2 import (
+from hgitaly.stub.repository_pb2 import (
     CreateRepositoryRequest,
     RepositoryExistsRequest,
 )
 from hgitaly.stub.shared_pb2 import Repository
-from hgitaly.stub.repository_service_pb2_grpc import RepositoryServiceStub
+from hgitaly.stub.repository_pb2_grpc import RepositoryServiceStub
+from hgitaly.testing.storage import (
+    DEFAULT_STORAGE_NAME,
+    git_repo_path,
+)
 
 from . import skip_comparison_tests
 if skip_comparison_tests():  # pragma no cover
@@ -27,12 +31,11 @@ def test_gitaly_channel(gitaly_channel, server_repos_root):
     channel = gitaly_channel
     repo_stub = RepositoryServiceStub(channel)
 
-    repo = Repository(relative_path='foo.git', storage_name='default')
+    relpath = 'foo.git'
+    repo = Repository(relative_path=relpath, storage_name=DEFAULT_STORAGE_NAME)
     repo_stub.CreateRepository(CreateRepositoryRequest(repository=repo))
 
-    # TODO should have a test util for that:
-    # join(server_repos_root, repo)
-    git_path = server_repos_root / 'default' / 'foo.git'
+    git_path = git_repo_path(server_repos_root, relpath)
     assert git_path.is_dir()
     assert (git_path / 'HEAD').exists
 
