@@ -4,6 +4,7 @@
 # GNU General Public License version 2 or any later version.
 #
 # SPDX-License-Identifier: GPL-2.0-or-later
+import os
 from pathlib import Path
 from mercurial import (
     node as node_mod,
@@ -28,6 +29,7 @@ from hgitaly.stub.repository_pb2 import (
     FindLicenseRequest,
     FindLicenseResponse,
     FindMergeBaseRequest,
+    HasLocalBranchesRequest,
     GetArchiveRequest,
     ObjectFormatRequest,
     RemoveRepositoryRequest,
@@ -194,10 +196,14 @@ def test_compare_create_repository(
         assert 'exists already' in exc.details()
 
 
-def test_remove_repository(gitaly_comparison, server_repos_root):
-    fixture = gitaly_comparison
+@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
+def test_remove_repository(gitaly_rhgitaly_comparison,
+                           hg_server,
+                           server_repos_root):
+    fixture = gitaly_rhgitaly_comparison
     grpc_repo = fixture.gitaly_repo
     rpc_helper = fixture.rpc_helper(
+        hg_server=hg_server,
         stub_cls=RepositoryServiceStub,
         method_name='RemoveRepository',
         request_cls=RemoveRepositoryRequest,
@@ -209,8 +215,25 @@ def test_remove_repository(gitaly_comparison, server_repos_root):
                           repository=Repository(storage_name='unknown',
                                                 relative_path='/some/path'))
     assert_compare_errors(
+        same_details=False,  # RHGitaly displays the path
         repository=Repository(storage_name=grpc_repo.storage_name,
                               relative_path='no/such/path'))
+    if hg_server != 'rhgitaly':
+        return
+
+    # specific RHGitaly testing, see heptapod#2275
+    workdir = (server_repos_root / 'default/+hgitaly/tmp/workdirs'
+               / fixture.hgitaly_repo.relative_path / '0')
+    os.makedirs(workdir)
+    (workdir / 'foo').write_bytes(b'some foo')
+    aux_git = (server_repos_root / 'default/+hgitaly/hg-git'
+               / fixture.gitaly_repo.relative_path)
+    os.makedirs(aux_git)
+    (aux_git / 'HEAD').write_bytes(b'ref: refs/heads/branch/default\n')
+
+    rpc_helper.rpc('hg')
+    assert not workdir.exists()
+    assert not aux_git.exists()
 
 
 @parametrize('hg_server', ('hgitaly', 'rhgitaly'))
@@ -336,7 +359,7 @@ def test_has_local_branches(gitaly_rhgitaly_comparison,
         hg_server=hg_server,
         stub_cls=RepositoryServiceStub,
         method_name='HasLocalBranches',
-        request_cls=RemoveRepositoryRequest,
+        request_cls=HasLocalBranchesRequest,
     )
     assert_compare = rpc_helper.assert_compare
     assert_compare_errors = rpc_helper.assert_compare_errors
